"""
Author: N. Abrate.

File: ndcovariance.py

Description: Class to read covariance/correlation matrices from NJOY file.
"""

import os
import numpy as np
import pandas as pd
import itertools as it
import matplotlib.pyplot as plt
from matplotlib import colors, ticker
from collections import OrderedDict
from mpl_toolkits.axes_grid1 import make_axes_locatable


class ndcovariance:
    """
    Define instance with nuclide-wise covariance and correlation matrices.

    Attributes
    ----------
    cov : dict
        Dictionary containing zais-wise covariance matrix
    cor : dict
        Dictionary containing zais-wise correlation matrix
    zais : list
        Nuclide identifier strings
    zais : list
        Nuclide identifier integers plus final zero
    perts : dict
        Perturbation reactions contained inside covariance/correlation matrix
    energies : numpy.ndarray
        Energy grid covariance is evaluated with respect to
    dopplerTemps : list
        List of temperatures used to evaluate covariance matrices
    nEne : int
        Number of energy groups.

    Methods
    -------
    read_cov_corr(path, NG, MT):
        Assemble bloc-wise covariance and correlation matrices for MT reactions
    getdiagblock(self, MT, ZAID, corr=False):
        Return sub-matrix diagonal block from complete covariance matrix
    parse_NJOY_covmat(path, NG, MT):
        Parse covariance matrix produced by NJOY
    cov2cor(C):
        Convert correlation matrix to covariance
    symmetrise(C):
        Make input upper triangular matrix symmetric.
    """

    def __init__(self, path, NG, MT, zais, temperatures=None, verbosity=False):
        # define periodic table
        ZtoAS, AStoZ = ASZ_periodic_table()
        # read energy grid and convert in MeV
        energies = np.loadtxt(os.path.join(path, 'E'))*1e-6

        # make MT nomenclature consistent with serpentTools
        MT.sort()  # sort (ascending)
        perts = {}
        for imt, mt in enumerate(MT):
            if mt == 1:
                perts['total xs'] = imt
            else:
                perts['mt %g xs' % mt] = imt

        perts = OrderedDict(perts.items())

        if temperatures is None:
            temperatures = [293]

        # define dict
        covdict = {}
        cordict = {}
        ZAIS, ZAID = [], []
        # loop over ZA inside ZAID
        for za in zais:
            # parse ZAid
            if isinstance(za, int):
                strZA = str(za)
                Z = strZA[0:2]
                AS = ZtoAS[int(Z)]
                A = strZA[2:-1]
                ZAIS.append('%s-%s' %(AS, A))
                ZAID.append(za)
            else:
                AS, A = za.split('-')
                ZAIS.append(za)
                Z = AStoZ[AS]
                za = int('%s%s0' %(Z, A))
                ZAID.append(za)

            path_za = os.path.join(path, str(za))

            covdict[za] = {}
            cordict[za] = {}
            for t in temperatures:
                path_za_t = os.path.join(path_za, "%dK" % t)
                # read covariance and correlation matrices from NJOY files
                cov, cor = ndcovariance.read_cov_corr(path_za_t, NG, MT,
                                                      verbosity=verbosity)
                # assign key and value
                covdict[za][t] = cov
                cordict[za][t] = cor

        self.cov = covdict
        self.cor = cordict
        self.zais = ZAIS
        self.zaid = ZAID
        self.perts = perts
        self.energies = energies
        self.dopplerTemps = temperatures
        self.nEne = len(self.energies)-1

    def read_cov_corr(path, NG, MT, verbosity=False):
        """
        Assemble bloc-wise covariance and correlation matrices for MT reactions.

        Parameters
        ----------
        path : str
            Covariance matrices path.
        NG : int
            Number of energy groups.
        MT : list[int]
            Elementary nuclear reactions.

        Returns
        -------
        covmat : numpy.ndarray
            Covariance matrix.
        cormat : numpy.ndarray
            Correlation matrix.
        """
        NMT = len(MT)
        # allocate nested lists (matrix blocks)
        covlst = [[] for i in it.repeat(None, NMT)]

        # build MT matrix
        for iMT, MT1 in enumerate(MT):  # rows
            for jMT, MT2 in enumerate(MT):  # columns
                # fill upper diagonal blocks
                if iMT <= jMT:
                    # parse matrices from NJOY output
                    CV = ndcovariance.parse_NJOY_covmat(path, NG, [MT1, MT2],
                                                        verbosity=verbosity)
                    covlst[iMT].append(CV)
                else:
                    covlst[iMT].append(covlst[jMT][iMT].T)

        # convert blocks to ndarray
        covmat = np.block(covlst)
        cormat = ndcovariance.cov2cor(covmat)

        return covmat, cormat

    def getdiagblock(self, MT, ZAID, T, corr=False):
        """
        Return sub-matrix diagonal block from complete covariance matrix.

        Parameters
        ----------
        MT : int
            MT reaction of the desired block.

        Returns
        -------
        CV: numpy.ndarray
            Covariance matrix block
        CR: numpy.ndarray
            Correlation matrix block
        """
        if type(MT) is not int:
            raise TypeError('MT reaction number must be integer!')

        # convert into string (key of dict)
        if MT == 1:
            MT = "total xs"
        else:
            MT = "mt %g xs" % MT

        idx = self.nEne*(self.perts[MT])
        # get block
        CV = self.cov[ZAID][T][0+idx:self.nEne+idx, 0+idx:self.nEne+idx]

        # compute correlation matrix, if needed
        if corr is True:
            # FIXME: why don I used 160 with self.cor?
            CR = ndcovariance.cov2cor(CV)
        else:
            CR = None

        return CV, CR

    def parse_NJOY_covmat(path, NG, MT, verbosity=False):
        """
        Parse covariance matrix produced by NJOY.

        Parameters
        ----------
        path : str
            Covariance files path.
        NG : int
            Number of energy groups.
        MT : list[int, int]
            List of MT reactions.

        Raises
        ------
        FileNotFoundError
            DESCRIPTION.

        Returns
        -------
        C : numpy.ndarray
            Covariance matrix.
        """
        # preallocation
        CVM = np.empty((NG, NG))
        # compute rows per group with 8 columns max per row
        rowbatch = int(np.ceil(NG/8))

        try:
            # compose file name
            fname = "MT%gMT%g" % (MT[0], MT[1])
            dpath = os.path.join(path, fname)
            # read file and bring it to numpy format
            data = pd.read_fwf(dpath, widths=[10]*8, skiprows=4).to_numpy()
            # matrix transpose flag
            transpose = False
        except FileNotFoundError:
            # exchange MT position
            try:
                fname = "MT%gMT%g" % (MT[1], MT[0])
                dpath = os.path.join(path, fname)
                data = pd.read_fwf(dpath, widths=[10]*8, skiprows=4).to_numpy()
                # data = np.genfromtxt(path, missing_values=np.nan)
                transpose = True
            except FileNotFoundError:
                if verbosity is True:
                    print("Covariance matrix %s not found!" % dpath)
                # exit without processing
                return np.zeros((NG, NG))

        # reshape data
        blocksize = int(rowbatch*data.shape[1])
        for irow in range(0, int(data.shape[0]/rowbatch)):
            # reshape covmat row
            iS = irow*rowbatch  # starting row
            iE = iS+rowbatch  # ending row
            # reshape block in row-array
            row = data[iS:iE, :].reshape((1, blocksize))
            # fill covariance matrix
            CVM[irow, :] = row[~np.isnan(row)]

        # rotate to flip from high-to-low to low-to-high energy
        CVM = np.rot90(np.rot90(CVM))

        if transpose is True:
            CVM = CVM.T

        return CVM

    def cov2cor(C):
        """
        Convert correlation matrix to covariance.

        Parameters
        ----------
        C : numpy.ndarray
            Covariance matrix.

        Returns
        -------
        CRM : numpy.ndarray
            Correlation matrix.
        """
        # extract standard deviation
        std = np.sqrt(np.diag(C))
        # build variance matrix
        varmat = np.outer(std, std)
        # define correlation matrix
        CRM = C/varmat
        # squeeze out NaN due to 0 in C
        CRM[C == 0] = 0
        return CRM

    def symmetrise(C):
        """
        Make input upper triangular matrix symmetric.

        Parameters
        ----------
        C : numpy.ndarray
            Upper triangular matrix.

        Returns
        -------
        A : numpy.ndarray
            Symmetric matrix.
        """
        A = C + C.T - np.diag(C.diagonal())
        # check symmetrise works
        assert np.all(A == A.T)

        return A

    def writeSERPCOV(self, path=None):
        """
        Write covariance matrices in a format compatible with Serpent 2.

        Parameters
        ----------
        path : str
            Path where files are saved.

        Returns
        -------
        None.
        """
        NG = len(self.energies)-1

        if path is None:
            path = os.getcwd()

        # get MT integer identifier
        MT = []
        for mt in self.perts:

            if "total" not in mt:
                MT.append(mt.split(" ")[1])

            else:
                MT.append('1')

        MTs = "_".join(MT)

        for za in self.zais:
            fname = os.path.join(path, "cov_%s_%s.txt" % (za, MTs))

            with open(fname, 'w') as f:
                # write file header
                f.write('%g\n' % NG)

                EneBound = ['%.10e' % e for e in self.energies]
                f.write(' '.join(EneBound))

                f.write('\n%g' % len(MT)**2)

                # write MT-wise covariance matrices
                for iMT, MT1 in enumerate(MT):  # rows

                    for jMT, MT2 in enumerate(MT):  # columns
                        f.write('\n%s %s %s %s\n' % (za, MT1, za, MT2))

                        idx = self.nEne*iMT
                        idy = self.nEne*jMT

                        CV = self.cov[za][0+idx:self.nEne+idx, 0+idy:self.nEne+idy]
                        CV = CV.flatten()
                        CV = ['%.10e' % el for el in CV]

                        f.write(''.join(CV))

    def plot(self, whichZAID, whichTemp=None, corr=True, logscale=True,
             figname=None, myax=None, thresh=None, title=None):

        """
        Plot covariance matrix.

        Parameters
        ----------
        figname : TYPE, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        None.
        """
        MTreactdict = MTdict()
        if isinstance(whichZAID, int):
            zadict = dict(zip(self.zaid, self.zais))
            ZAIS = zadict[whichZAID]
            ZAID = whichZAID
        else:
            zadict = dict(zip(self.zais, self.zaid))
            ZAID = zadict[whichZAID]
            ZAIS = whichZAID

        # select temperature
        if whichTemp is None:
            whichTemp = self.dopplerTemps[0]
        # get MT reaction numbers
        nMTs = len(self.perts.keys())
        MT = []
        for p in self.perts.keys():
            # get MT identifier
            if "total" not in p:
                MT.append(int(p.split(" ")[1]))
            else:
                MT = MT.append(1)

        MTs = [p for p in it.product(MT, repeat=2)]
        # get block
        if corr is False:
            M = self.cov[ZAID][whichTemp]
            if title is not None:
                title = '%d covariance matrix, %dG, T=%d K' % (ZAIS, self.nEne,
                                                               whichTemp)
        else:
            M = self.cor[ZAID][whichTemp]
            if title is not None:
                title = '%s correlation matrix, %dG, T=%d K' % (ZAIS, self.nEne,
                                                                whichTemp)
        linthresh = np.min(abs(M)[np.nonzero(M)])
        # define figure
        fig, axs = plt.subplots(nMTs, nMTs, figsize=(5, 5),
                                constrained_layout=True, sharex=True,
                                sharey=True)

        # fill subplot with submatrices
        nsk = self.nEne  # number of elements to be skipped
        row, col = 0, 0
        E = self.energies
        for nax, ax in enumerate(axs.flat):
            if nax != 0 and nax % nMTs == 0:
                row = row+1  # new row
                col = 0  # rewind column
            # parse submatrix
            subM = M[0+row*nsk:nsk*(1+row), 0+col*nsk:nsk*(1+col)]

            # apply threshold
            if thresh is not None:
                if isinstance(thresh, list):
                    thresh.sort()
                    subM[(subM > thresh[0]) & (subM < thresh[1])] = 0
                else:
                    if thresh > 0:
                        subM[subM > thresh] = 0
                    else:
                        subM[subM < thresh] = 0

            if logscale is True:
                im = ax.pcolormesh(E, E, subM, cmap='RdBu',
                                   norm=colors.SymLogNorm(vmax=M.max(),
                                                          vmin=M.min(),
                                                          linthresh=linthresh))
            else:
                im = ax.pcolormesh(subM, cmap='RdBu')

            ax.set_aspect('equal')
            ax.set_xscale('log')
            ax.set_yscale('log')
            MT1, MT2 = MTs[nax]
            # add labels
            if col == 0:
                ax.set_ylabel("%s" % MTreactdict[MT1])
            if row == nMTs-1:
                ax.set_xlabel("%s" % MTreactdict[MT2])
            # ax.title.set_text('%s-%s' % (, MTreactdict[MT2]))
            col = col + 1  # new column

        # add features
        # plt.subplots_adjust(hspace=0.1, wspace=0.1)
        if title is not None:
            fig.suptitle(title)
        cb = fig.colorbar(im, ax=axs, ticks=[-1, -1E-2, 0, 1E-2, 1]) # , format=ticker.LogFormatterMathtext()
        # tick_locator = ticker.MaxNLocator(nbins=5)
        # cb.locator = tick_locator
        # cb.update_ticks()

        fig.add_subplot(111, frameon=False)
        plt.tick_params(labelcolor='none', top=False, bottom=False,
                        left=False, right=False)

        fig.text(0.5, -0.05, 'E [MeV]', ha='center')
        fig.text(-0.05, 0.5, 'E [MeV]', va='center', rotation='vertical')
        # plt.subplots_adjust(hspace=0.1, wspace=0.1)
        if figname is not None:
            plt.savefig("%s.png" % figname, bbox_inches = 'tight')


def MTdict():
    """
    Define a dictionary with MT reaction numbers and descriptive string.

    Returns
    -------
    MTdict : dict
        key: MT number.
        value: descriptive string.
    """
    MTdict = {1: 'tot', 2: '(n,n)', 18: '(n, fiss.)', 102: '(n, phot.)',
              1800: '$\Sigma_f$', 10200: '$\Sigma_c$'}

    return MTdict


def ASZ_periodic_table():
    """
    Define a dictionary with atomic symbol and atomic number.

    Returns
    -------
    periodictable : dict
        key: atomic symbol AS.
        value: atomic number Z.
    """
    # define atomic symbols list
    AS = ['H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg',
          'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr',
          'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br',
          'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd',
          'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La',
          'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er',
          'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au',
          'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th',
          'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md',
          'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', 'Cn',
          'Nh', 'Fl', 'Mc', 'Lv', 'Ts', 'Og']
    # atomic number
    Z = np.arange(1, 119)
    # define dict
    ZtoAS = dict(zip(Z, AS))
    AStoZ = dict(zip(AS, Z))
    # return element dict
    return ZtoAS, AStoZ
