"""
Author: N. Abrate.

File: perturbedXS.py

Description: Class to read perturbed nuclear cross sections.
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import rc
from scipy.io import loadmat
from numpy.linalg import norm


class perturbedXS:
    """
    Apply sandiwch rule to GPT sensitivity coefficients.

    Attributes
    ----------
    E: numpy.ndarray
        XS energy grid
    nEne: int
        number of energy points
    Z : int
        atomic number
    A : int
        mass number
    lib : str
        library
    temp : int
        Doppler temperature evaluation
    source : str
        perturbation source
    MT : list
        list of MT reactions present in the snapshot matrix
    MTmap : dict
        snapshot matrix map
    snapshotmat : numpy.ndarray
        snapshot matrix
    refData : numpy.ndarray
        reference data evaluation (unperturbed)

    Methods
    -------
    show_recon_XS(self, MT, case, PODcoeff, PODbasis, figname=None, xlim=None):
        Plot data and POD-reconstructed data
    ReadMatlab(self, matname, mtlist=None):
        Read XS from .mat file
    """

    def __init__(self):
        return None

    def show_recon_XS(self, MT, case, PODcoeff, PODbasis, figname=None, xlim=None):
        """

        Plot data and POD-reconstructed data.

        Parameters
        ----------
        MT : int
            MT reaction to be interpoalted and plotted
        case : int
            Case number to identify perturbed cross section
        PODcoeff : numpy.ndarray
            POD reduced space coefficients
        PODbasis : numpy.ndarray
            POD basis functions
        figname : str
            Name for saving figure
        xlim : list
            List with two x-axis limits

        Returns
        -------
        None.
        """
        # take coordinates from the snapshot map
        iS, iE = self.MTmap[MT]

        # interpolate XS
        PODsnap = np.dot(PODbasis, PODcoeff[:, case])
        interpol = (1+PODsnap)[iS:iE+1, np.newaxis]*self.refData[iS:iE+1]
        refxs = (1+self.snapshotmat[iS:iE+1, case])[:, np.newaxis]*self.refData[iS:iE+1]

        plt.style.use('seaborn-whitegrid')
        rc('text', usetex=False)

        fig = plt.figure()

        lin1, = plt.loglog(self.E, refxs)
        lin2, = plt.loglog(self.E, interpol, '--')

        if xlim is not None:
            plt.xlim(xlim)

        handles = (lin1, lin2)
        labels = ("XS", "XS ROM")

        plt.xlabel("Energy [MeV]")
        plt.ylabel("MT%g [b]" % MT)

        legend_x = 1
        legend_y = 0.5
        plt.legend(handles, labels, bbox_to_anchor=(legend_x, legend_y),
                   loc='center left', ncol=1)

        abserr = norm(np.dot(PODbasis, PODcoeff[:, case]) -
                      self.snapshotmat[:, case])
        relerr = abserr/norm(self.snapshotmat[:, case])
        title = "Relative error: %.2f" % relerr + "%"
        plt.title(title)
        plt.show()

        if figname is not None:
            fig.savefig(figname, bbox_inches='tight', dpi=500)

    @classmethod
    def ReadMatlab(self, matname, mtlist=None):
        """
        Read XS from .mat file.

        Parameters
        ----------
        matname : str
            Matlab ".mat" file name
        mtlist : list
            List of MT reactions to be extracted

        Returns
        -------
        None
        """
        print("Reading .MAT file...")
        fmat = loadmat(matname, struct_as_record=False, squeeze_me=True)
        print("DONE")

        # pop all keys out except struct name
        keys = list(fmat.keys())
        structname = [k for k in keys if not k.startswith("_")]

        # retrive MATLAB struct content
        matstruct = fmat[structname[0]]

        # build XS snapshot matrix
        if mtlist is None:

            MT = matstruct.MT_list.sort()

        else:

            MT = [mt for mt in matstruct.MT_list if mt in mtlist]
            MT.sort()

        # compute number of energy points
        nEne = len(matstruct.E)

        # compute number of snapshots
        nSnap = matstruct.__dict__["MT%g" % MT[0]].shape[1]

        # build snapshot matrix
        Y = np.empty((len(MT)*nEne, nSnap))
        refData = np.empty((len(MT)*nEne, 1))
        N_old = 0
        MTmap = {}

        for mt in MT:
            mtfield = "MT%g" % mt
            mtfield_ref = "ref_MT%g" % mt
            iS = N_old
            iE = iS+nEne-1
            # snapshot
            Y[iS:iE+1, :] = matstruct.__dict__[mtfield]
            # reference data
            refData[iS:iE+1, :] = matstruct.__dict__[mtfield_ref][:, np.newaxis]
            # save MT positions
            MTmap[mt] = [iS, iE]
            N_old = iE+1

        # normalise deviations wrt reference data
        Y = np.subtract(Y, refData)/refData

        # transfer MATLAB attributes to self (rename)
        self.E = matstruct.E
        self.nEne = nEne
        self.Z = matstruct.Z
        self.A = matstruct.A
        self.lib = matstruct.nlib
        self.temp = matstruct.T
        self.source = matstruct.source
        self.MT = MT
        self.MTmap = MTmap
        self.snapshotmat = Y
        self.refData = refData


@staticmethod
def __checkname(fname):
    """
    Check extension of filename and add if it is missed.

    Parameters
    ----------
    fname : string
        Input filename (optional extension).

    Raises
    ------
    OSError
        -File extension is wrong. Only HDF5 can be parsed

    Returns
    -------
    fname : string
        Filename and proper extension.

    """
    lst = fname.split(".")

    if len(lst) == 1:
        # add extension
        fname = fname+".hdf5"
    else:
        if lst[-1] != "hdf5":
            lst[-1] = "hdf5"
            fname = ".".join(lst)
            raise OSError("File extension is wrong. Only HDF5 can be parsed")

    return fname
