"""
Author: N. Abrate.

File: h5.py

Description: Utility to read/write objects from/to HDF5 files.
"""
import os
import h5py
import chardet
import numpy as np
from numpy import string_, ndarray, array, asarray
from numpy import int8, int16, int32, int64, float16, float32, float64


class read():
    """
    Class that reads a .hdf5/.h5 file and stores it into an object.

    h5name : str
        HDF5 file name that is parsed.
    """

    _float_types = (float, float16, float32, float64)
    _int_types = (int, int8, int16, int32, int64)
    _arr_types = (ndarray, list)
    _types = (*_int_types, *_float_types, *_arr_types, str,
              h5py._hl.dataset.Dataset)

    def __init__(self, h5name, metadata=True):
        """
        Read .hdf5/.h5 formatted files and store in object.

        Parameters
        ----------
        h5name : string
            File where data are stored.
        metadata : bool, optional
            bool to choose if metadata are parsed or ignored. Default is
            ``True``.

        Returns
        -------
        ``None``
        """
        # open hdf5 file
        h5name = _checkname(h5name)
        fh5 = h5py.File(h5name, "r")

        # FIXME: add selective reading (maybe specifying group paths)

        # extract groups
        for gro in fh5.keys():

            self.__dict__[gro] = {}
            dic = self.__dict__[gro]

            # get attributes
            if metadata is True:
                self.__dict__[gro]['%s_metadata' % gro] = read._attrs2dict(fh5[gro])

            # get sub-groups and datasets
            for k, v in fh5[gro].items():

                try:
                    pytype = v.attrs['pytype']

                except KeyError:
                    pytype = None

                dic[k] = read._h52py(v, pytype)

                if metadata is True:
                    attrs = read._attrs2dict(v)

                    try:
                        if attrs != {}:
                            dic[k]['%s_metadata' % str(k)] = attrs

                    except IndexError:
                        if attrs != {}:
                            dic['%s_metadata' % str(k)] = attrs

        fh5.close()
        self.h5name = h5name

    def _h52py(item, pytype=None):
        """
        Convert item from HDF5 file to the closest Python-3 type.

        Parameters
        ----------
        item : string, int, float, numpy.array, dict, list
            Dataset/subgroup read from HDF5 file.
        pytype : type
            Python type assigned as attribute to the Dataset/subgroup. Default
            is ``None`` if the attribute does not exist.

        Returns
        -------
        val : int
            Item converted in Python type (str, int, float, numpy.array, dict,
            list).
        """
        if pytype is None:

            try:
                pytype = item.dtype.name

            except AttributeError:

                if isinstance(item, str):
                    val = item
                else:
                    val = read._h52dict(item)

                return val

        if isinstance(item, read._types):

            if isinstance(item, read._arr_types) or item.size > 1:
                val = array(item)

            elif (isinstance(item, read._float_types) or item.dtype in read._float_types) and len(item.shape) == 0:
                val = array(item)
                val = float(val)

            elif (isinstance(item, read._int_types) or item.dtype in read._int_types) and len(item.shape) == 0:
                val = array(item)
                val = int(val)

            elif pytype == 'str':
                val = item

            elif pytype == b'str':

                try:
                    enc = item.attrs['encoding'].decode('ascii')

                except KeyError:
                    enc = 'ascii'
                    print('Warning: %s of type b''str'' read with ''ascii'' encoding')

                val = item.value.decode(enc)

            else:
                raise TypeError('Cannot read data %s!' % item)

        elif isinstance(item, h5py._hl.group.Group):

            try:
                if pytype == b'dict':
                    val = read._h52dict(item)

            except TypeError:
                raise TypeError('Cannot read data %s!' % item)

        else:
            raise TypeError('Cannot read data %s!' % item)

        return val

    def _h52dict(path):
        """
        Convert group/subgroup/dataset to dictionaries.

        Parameters
        ----------
        path : str
            Path to dataset in the HDF5 file.

        Returns
        -------
        dic: dict
            group/subgroup/dataset converted to dictionary.
        """
        # FIXME: try parallel computation to speed up. Too slow!
        dic = {}
        for key, it in path.items():

            if isinstance(it, h5py._hl.dataset.Dataset):
                dic[key] = it[()]

            elif isinstance(it, h5py._hl.group.Group):
                dic[key] = read._h52dict(it)

        return dic

    def _attrs2dict(path):
        """
        Parse attributes from group/subgroup/dataset.

        Convert group/subgroup/dataset to dictionaries.

        Parameters
        ----------
        path : str
            Path to dataset in the HDF5 file.

        Returns
        -------
        dic: dict
            attributes converted to dictionary.
        """
        dic = {}
        for k, v in path.attrs.items():
            dic[k] = read._h52py(v)

        return dic


def readXS(fname, serpkeys, NG, mean=True, pert=True, uni=None):
    """
    Read point-wise cross sections.

    Parameters
    ----------
    fname : str
        Name of the HDF5 file with or without extension.
    serpkeys : list[strings]
        Dict keys of the data read with ``serpentTools``.
    mean : bool, optional
        ``True`` to get the mean, ``False`` to get the rsd.
    pert : bool, optional
        Default is ``True``.
    uni : str, optional
        Universe to where XS are computed. Default is ``None``.
    Returns
    -------
    SM : ndarray
        snapshot matrix of data generated by Serpent 2
    parameters : tuple(arrays)
        parameters employed
    """
    # check which statistical quantity is desired
    if mean is True:
        stats = "mean"
    else:
        stats = "rsd"

    fname = _checkname(fname)
    f = h5py.File(fname, "r")

    # extract group universes
    if uni is None:
        universes = _getuniname(f, pert=pert)
    else:
        universes = uni

    # extract parameters combination
    SM, parameters = {}, {}
    if pert is True:
        # store data in snapshot matrix
        for uni in universes:  # loop over dict keys

            start, end = _parsepertkeys(list(f[uni].keys()))
            # parameters, snapshots
            n, m = (NG*(len(serpkeys)-1)+NG**2, int(end)-int(start)+1)
            U = np.nan*np.zeros((n, m))

            for ik, k in enumerate(range(int(start), int(end)+1)):

                path = "/".join((uni, str(k), stats))

                # collect data with list comprehension
                extractedata = [np.array(f.get(path + "/" + key))
                                for key in serpkeys if key.startswith('inf')]
                # flatten the matrix in a column array
                col = np.concatenate(extractedata)

                # assign row to column matrix
                U[:, ik] = col

            SM[uni] = U
            parameters[uni] = array(f[uni]['parametersgrid'])

    else:  # single universe

        for uni in universes:  # loop over dict universes
            SM[uni] = f[uni][stats]

    return SM, parameters


def write(obj, h5name, attrs=None, chunks=True, compression=True):
    """
    Write object in .h5 format.

    Parameters
    ----------
    obj : object or iterable
        class object to be saved in HDF5.
    h5name : string
        File where data are stored.
    attrs : dict
        Dictionary attributes that can be assigned for further description
        (default is ``None``).
    chunks : bool, optional
        Flag to save HDF5 in chunks.
    compression : bool, optional
        Flag to save compressed HDF5.

    Returns
    -------
    ``None``
    """
    # make input an iterable
    if isinstance(obj, list) is not True:
        if isinstance(obj, (object, ndarray)):
            obj = {obj, }

    if attrs is not None:
        if isinstance(attrs, list) is False:
            attrs = [attrs]
    else:
        attrs = [None]

    # open hdf5 file
    h5name = _checkname(h5name)
    fh5 = _wopen(h5name)

    if fh5 == -1:
        print("File not overwritten.")

    # profiling
    print("Writing object in HDF5 file...")

    # loop over object attributes
    for i, (ob, at) in enumerate(zip(obj, attrs)):
        # store lines to be found in a list
        if at is not None:
            fid = at['grp_name']

        else:
            fid = 'grp%g' % i
        # create group with output id
        fh5.create_group(fid)

        # store input dict as 1st group metadata
        if at is not None:
            for k, v in at.items():

                if k != 'grp_name':
                    fh5[fid].attrs[k] = v

            fh5[fid].attrs['pytype'] = str(type(ob))

        # loop over object attributes
        if isinstance(ob, (dict)):
            mydict = ob
        else:
            mydict = ob.__dict__

        for k2, v2 in mydict.items():

            if isinstance(v2, (dict)):
                # create sub-group
                dict2h5(fh5[fid], k2, v2)

            else:  # create dataset

                if isinstance(v2, (str)):
                    v2 = string_(v2)
                    fh5[fid].create_dataset(k2, data=v2)

                    fh5[fid][k2].attrs['pytype'] = string_('str')
                    # assign original encoding
                    enc = chardet.detect(v2).get('encoding')
                    fh5[fid][k2].attrs['encoding'] = string_(enc)

                else:
                    if compression is True:
                        try:
                            fh5[fid].create_dataset(k2, data=v2,
                                                    compression="gzip",
                                                    compression_opts=4,
                                                    chunks=True)
                        except TypeError:
                            if isinstance(v2, (list)):
                                if isinstance(v2[0], str):
                                    v2 = string_(v2)
                            fh5[fid].create_dataset(k2, data=v2)
                    else:
                        if isinstance(v2, list):
                            if isinstance(v2[0], str):
                                v2 = string_(v2)
                        fh5[fid].create_dataset(k2, data=v2)

    fh5.close()
    print("DONE \n")


def dict2h5(h5grp, path, dic):
    """
    Recursively write dict to HDF5 format.

    Parameters
    ----------
    h5grp : str
        HDF5 file object.
    path : str
        Group name inside HDF5 file.
    dic : dict
        Dictionary to be saved in HDF5 format.

    Raises
    ------
    ValueError
        If the dict contains types different from ndarray, int64, float64,
        bytes, int, string, list or dict, the method raises the error
        "Cannot save XXX type with dict2h5".

    Returns
    -------
    ``None``.

    """
    path = str(path) if isinstance(path, str) is False else path
    for key, item in dic.items():

        if isinstance(key, (int64, float64, int)):
            key = str(key)

        if isinstance(item, (ndarray, int64, float64, bytes, int)):
            h5grp['/'.join([path, key])] = item

        elif isinstance(item, str):
            h5grp['/'.join([path, key])] = string_(item)
            h5grp['/'.join([path, key])].attrs['pytype'] = string_('str')

        elif isinstance(item, list):
            h5grp['/'.join([path, key])] = asarray(item)
            h5grp['/'.join([path, key])].attrs['pytype'] = string_('list')

        elif isinstance(item, dict):
            dict2h5(h5grp, '/'.join([path, key]), item)

        else:
            raise ValueError('Cannot save %s type with dict2h5' % type(item))

    h5grp[path].attrs['pytype'] = string_('dict')


def _checkname(fname):
    """
    Check file extension.

    Parameters
    ----------
    fname : str
        Input filename (optional extension).

    Raises
    ------
    OSError
        -File extension is wrong. Only HDF5 can be parsed

    Returns
    -------
    ``None``.
    """
    lst = fname.split(".")

    if len(lst) == 1:
        raise OSError("File extension is missing. Only HDF5 can be read/written!")

    else:

        if lst[-1] != "hdf5" and lst[-1] != "h5":
            raise OSError("File extension is wrong. Only HDF5 can be read/written!")

    return fname


def _wopen(h5name):
    """
    Open the hdf5 file "h5name.hdf5" in "append" mode.

    Parameters
    ----------
    h5name : str
        File name

    Returns
    -------
    fh5 : object
        h5py object

    """
    if os.path.isfile(h5name):
        print("File exists. Overwriting?")
        ans = input()

    else:  # if file do not exist, it creates it
        ans = "create"

    if ans in ['yes', 'y', 'Yes', 'Y', 'YES']:
        os.remove(h5name)
        # open file in write mode
        fh5 = h5py.File(h5name, "a")

        return fh5

    elif ans == "create":
        # create file in append mode
        fh5 = h5py.File(h5name, "a")

        return fh5

    else:  # if answer is not positive, nothing is done
        return -1


def _getuniname(f, pert):
    """


    Parameters
    ----------
    f : TYPE
        DESCRIPTION.
    pert : TYPE
        DESCRIPTION.

    Returns
    -------
    univ : TYPE
        DESCRIPTION.

    """
    # parse universes
    univ = []
    univapp = univ.append
    keys = list(f.keys())
    for k in keys:

        data = f[k]

        if pert is True:
            start, end = _parsepertkeys(list(f[k].keys()))
            data = data[start]

        try:
            if 'mean' in data.keys():  # it is universe
                univapp(k)
            else:
                continue  # it is not a universe
        except AttributeError:
            continue

    return univ


def _parsepertkeys(lst):
    """
    Parse starting and ending perturbation indeces.

    Parameters
    ----------
    s : string
        Input string.

    Returns
    -------
    bool
        True is the string is an integer, False otherwise.
    """
    start, end = None, None
    idx = [i for i in lst if i.isdigit()]
    idx.sort(key=lambda x: int(''.join(filter(str.isdigit, x))))
    start, end = idx[0], idx[-1]

    return start, end


def _isint(s):
    """
    Check if input is integer.

    Parameters
    ----------
    s : string
        Input string.

    Returns
    -------
    bool
        True is the string is an integer, False otherwise.

    """
    try:
        int(s)
        return True
    except ValueError:
        return False
