"""
author: N. Abrate.

file: NDUQ.py

description: GPT and XGPT classes to perform uncertainty quantification for
            Serpent-2 output responses.
"""
import numpy as np
from numpy import trapz
from uncertainties import ufloat
from copy import deepcopy
from uncertainties import unumpy as unp
from scipy.interpolate import interp1d
from collections import OrderedDict


class ndGPT:
    """
    Apply sandwich rule to GPT sensitivity coefficients.

    Attributes
    ----------
    sensperts : dict
        Ordered dictionary of reactions that have been perturbed in
        sensitivity calculations, e.g 'mt 18 xs'
    covperts : dict
        Ordered dictionary of reactions that have been perturbed stored in
        covariance matrix, e.g 'mt 18 xs'
    respRVAR : numpy.ndarray
        Response relative variance +/- statistical relative uncertainty
    respRSD : numpy.ndarray
        Response relative standard deviation +/- statistical relative
        uncertainty
    respRVAR_tot : numpy.ndarray
        Total response relative variance +/- statistical relative uncertainty
    respRSD_tot : numpy.ndarray
        Total response relative standard deviation +/- statistical relative
        uncertainty

    Methods
    -------
    None

    """

    def __init__(self, sensT, covmat, matcomp, sigma=2, Tf=True,
                 verbosity=False):
        """
        Initialise class.

        Parameters
        ----------
        sensT: dict
            Dictionary with sensitivity objects created reading Serpent-2
            "_sens0.m" object with "serpentTools" module. The keys are the
            temperature couples used to evaluate the sensivity coefficients.
        covmat: object
            Covariance matrix object created using "ndcovariance" module.
        matcomp: dict
            Dictionary matching the nuclides considered for the sensitivity
            study to each material considered.
        sigma: confidence interval required for statistical error
                      propagation.

        Returns
        -------
        None
        """
        # make MT nomenclature consistent
        perts = covmat.perts

        nMat = len(matcomp.keys())

        respRSD = {}
        respRVAR = {}
        respRSD_tot = {}
        respRVAR_tot = {}
        self.materials = OrderedDict()

        for temp in sensT.keys():
            # get temperature couple
            sens = sensT[temp]
            respRVAR[temp] = {}
            respRSD[temp] = {}
            respRVAR_tot[temp] = {}
            respRSD_tot[temp] = {}
            # loop over responses
            for resp in sens.sensitivities.keys():

                rvar = np.full((nMat, sens.nZai, sens.nPert, 2), None)
                rsd = np.full((nMat, sens.nZai, sens.nPert, 2), None)

                rvar_tot = np.full((nMat, sens.nZai, 2), None)
                rsd_tot = np.full((nMat, sens.nZai, 2), None)

                # loop over user selected materials
                for imat, mat in enumerate(matcomp.keys()):
                    self.materials[mat] = imat
                    jmat = sens.materials[mat]
                    # loop over nuclides
                    for iza, za in enumerate(sens.zais):
                        # check nuclide presence in material
                        if za not in matcomp[mat]:
                            continue

                        # allocate array for all sensitivities
                        ST = np.full((len(perts)*sens.nEne, 2), None)
                        idx = 0

                        # loop over MT reactions
                        for imt, mt in enumerate(sens.perts):

                            # check za in covariance object
                            if za not in covmat.zaid:
                                if verbosity is True:
                                    print("%d covariance matrices not " % mt
                                          + "available!")
                                continue

                            # check mt in covariance object
                            if mt not in perts.keys():
                                if verbosity is True:
                                    print("%s covariance matrix " % mt +
                                          "not available!")
                                continue

                            # get MT identifier
                            if "total" not in mt:
                                MT = int(mt.split(" ")[1])
                            else:
                                MT = 1

                            # parse temperature for covariance
                            T = temp[0] if Tf is True else temp[1]
                            covtemps = np.asarray(list(covmat.cov[za].keys()))
                            idT = (np.abs(covtemps - T)).argmin()

                            # get nuclide-wise, mt-wise covariance block
                            C, CR = covmat.getdiagblock(MT, za, covtemps[idT],
                                                        corr=False)

                            # slice to get energy-wise sensitivity coeffs
                            S = sens.sensitivities[resp][jmat, iza, imt, :, :]

                            # store in total sensitivity array
                            ST[0+idx*sens.nEne:sens.nEne+idx*sens.nEne, :] = S
                            idx = idx+1

                            # convert S into ufloat type
                            S = np2unp(S[:, 0], sigma*S[:, 1])

                            # sandwich rule (relative variance with stat. unc.)
                            sandwhich = np.dot(S.T, np.dot(C, S))
                            rvar[imat, iza, imt, :] = sandwhich.n, sandwhich.s

                            sand_sqrt = unp.sqrt(sandwhich).item(0)
                            rsd[imat, iza, imt, :] = sand_sqrt.n, sand_sqrt.s

                        # convert S into ufloat type
                        ST = np2unp(ST[:, 0], sigma*ST[:, 1])

                        # MT cross-corr sandwich rule (rel.var. with stat.unc.)
                        CT = covmat.cov[za][covtemps[idT]]
                        sandwhich_tot = np.dot(ST.T, np.dot(CT, ST))
                        rvar_tot[imat, iza, :] = sandwhich_tot.n, sandwhich_tot.s
                        sand_tot_sqrt = unp.sqrt(sandwhich_tot).item(0)
                        rsd_tot[imat, iza, :] = sand_tot_sqrt.n, sand_tot_sqrt.s

                respRVAR[temp][resp] = rvar
                respRSD[temp][resp] = rsd

                respRVAR_tot[temp][resp] = rvar_tot
                respRSD_tot[temp][resp] = rsd_tot

        self.covperts = covmat.perts
        self.respRVAR = respRVAR
        self.respRSD = respRSD
        self.respRVAR_tot = respRVAR_tot
        self.respRSD_tot = respRSD_tot
        # return other attributes
        self.zaid = sens.zais  # what in sens is zais here is zaid
        self.zais = dict(zip(covmat.zais, covmat.zaid))
        self.mts = sens.perts

    def get(self, attr, resp, mat, zai, temp, mt=None):
        """
        Get mean and standard deviation from XGPT object.

        Parameters
        ----------
        attr : string
            Object attribute. It may be 'respRVAR', 'respRSD', 'respRVAR_tot'
            or 'respRSD_tot'.
        mat : string
            Material identifier.
        zai : string
            Z-A identifier.
        temp : tuple
            Couple of fuel and coolant temperatures used to evaluate the
            responses.
        mt : string, optional
            MT reaction. This is only needed for 'respRVAR' and 'respRSD'
            attributes. The default is None.

        Returns
        -------
        vals : numpy.array
            Mean and standard deviation required.
        """
        vals = []
        iMat = self.materials[mat]
        try:
            iZai = self.zaid[zai]
        except KeyError:
            iZai = self.zaid[self.zais[zai]]

        if attr.endswith('_tot') is False:

            if mt is None:

                raise ValueError('Specify MT numbers!')

            else:

                iMT = self.mts[mt]

            for k in self.__dict__[attr][temp].keys():

                if k.startswith(resp):

                    vals.append(self.__dict__[attr][temp][k][iMat, iZai, iMT, :])

        else:

            for k in self.__dict__[attr][temp].keys():

                if k.startswith(resp):

                    vals.append(self.__dict__[attr][temp][k][iMat, iZai, :])

        return np.asarray(vals, dtype=float)


class ndXGPT:
    """
    Apply sandwich rule to XGPT sensitivity coefficients.

    Attributes
    ----------
    sensperts : dict
        Ordered dictionary of reactions that have been perturbed in
        sensitivity calculations, e.g 'mt 18 xs'
    respRVAR : numpy.ndarray
        Response relative variance +/- statistical relative uncertainty
    respRSD : numpy.ndarray
        Response relative standard deviation +/- statistical relative
        uncertainty
    respRVAR_tot : numpy.ndarray
        Total response relative variance +/- statistical relative uncertainty
    respRSD_tot : numpy.ndarray
        Total response relative standard deviation +/- statistical relative
        uncertainty

    Methods
    -------
    getInfo(perts):
        Get mat, ZAID, MT and perturbation number from sens.perts list
    """

    def __init__(self, indsensT, RObasis, ROmap, matcomp, reduction,
                 sigma=2, dirsensT=None, verbosity=False, zaid=None,
                 nBasFunc=None, noUQ=True):
        """
        Initialise class.

        Parameters
        ----------
        indsensT: object
            Dictionary with sensitivity objects created reading Serpent-2
            "_sens0.m" object with "serpentTools" module. The keys are the
            temperature couples used to evaluate the sensivity coefficients.
        RObasis: object
            Reduced Order basis function object generated by "POD" module.
        matcomp: dict
            Dictionary matching the nuclides considered for the sensitivity
            study
                       to each material considered.
        ROmap: dict
            Dictionary mapping the basis functions to the nuclides.
        reduction: str
            Kind of data reduction employed: 'pod' for Proper Orthogonal
            Decomposition performed on the snapshot matrix, 'evd' for
            Eigenvalue Decomposition performed on the nuclide covariance
            matrix.
        sigma: confidence interval required for statistical error
                      propagation.

        Returns
        -------
        None
        """
        for temp in indsensT.keys():
            mats, zais, mts, SensMap = ndXGPT.getInfo(indsensT[temp].perts)
            zais = zais if zaid is None else [zaid]
            break

        nZai = len(zais) if zaid is None else 1
        nMT = len(mts)
        nMat = len(mats)

        respRSD = {}
        respRVAR = {}
        respRSD_tot = {}
        respRVAR_tot = {}
        for temp in indsensT.keys():
            # get temperature couple
            indsens = indsensT[temp]
            respRVAR[temp] = {}
            respRSD[temp] = {}
            respRVAR_tot[temp] = {}
            respRSD_tot[temp] = {}
            # loop over responses
            for resp in indsens.sensitivities.keys():

                rvar = np.full((nMat, nZai, nMT, 2), None)
                rsd = np.full((nMat, nZai, nMT, 2), None)

                rvar_tot = np.full((nMat, indsens.nZai, 2), None)
                rsd_tot = np.full((nMat, indsens.nZai, 2), None)

                # loop over materials
                for imat, mat in enumerate(mats):
                    # loop over nuclides
                    for iza, za in enumerate(zais):

                        if za not in matcomp[mat]:
                            continue

                        # len equal for all mts inside mat inside za
                        if nBasFunc is None:
                            nBasFunc = len(SensMap[mat][za][mts[0]])

                        # get Reduced Order coefficients
                        if reduction == 'pod':
                            C = np.cov(RObasis.coeffs[0:nBasFunc])

                        else: # WATCH OUT: np.diag(np.sqrt(RObasis.eigv[0:nBasFunc]))
                            C = np.diag((RObasis.eigv[0:nBasFunc]))

                        for imt, mt in enumerate(mts):

                            if za not in ROmap.keys():
                                if verbosity is True:
                                    print("%s basis functions not available!"
                                          % za)
                                continue

                            if mt not in ROmap[za].keys():
                                if verbosity is True:
                                    print("%s covariance matrix not available!"
                                          % mt)
                                continue

                            # get energy integrated sensitivity coeffs
                            pos = SensMap[mat][za][mt]
                            iS, iE = min(pos), max(pos)
                            eintsens = indsens.energyIntegratedSens[resp]
                            indS = eintsens[0, 0, iS:iE+1, :]
                            if dirsensT is None:
                                dirS = 0*eintsens[0, 0, iS:iE+1, :]
                            else:
                                eintdirsens = dirsensT[temp].energyIntegratedSens[resp]
                                dirS = eintdirsens[0, 0, iS:iE+1, :]
                            # convert S into ufloat type
                            if noUQ is True:
                                indS = indS[0:nBasFunc, 0]
                                dirS = dirS[0:nBasFunc, 0]
                            else:
                                indS = np2unp(indS[0:nBasFunc, 0],
                                              sigma*indS[0:nBasFunc, 1])
                                dirS = np2unp(dirS[0:nBasFunc, 0],
                                              sigma*dirS[0:nBasFunc, 1])
                            S = indS+dirS

                            # sandwich rule (relative variance with stat. unc.)
                            sandwhich = np.dot(S.T, np.dot(C, S))

                            if noUQ is True:
                                rvar[imat, iza, imt, :] = sandwhich, sandwhich*10
                                sand_sqrt = np.sqrt(sandwhich).item(0)
                                rsd[imat, iza, imt, :] = sand_sqrt, sand_sqrt*10
                            else:
                                rvar[imat, iza, imt, :] = sandwhich.n, sandwhich.s
                                sand_sqrt = unp.sqrt(sandwhich).item(0)
                                rsd[imat, iza, imt, :] = sand_sqrt.n, sand_sqrt.s

                            if rvar_tot[imat, iza, 0] is None:
                                if noUQ is True:
                                    rvar_tot[imat, iza, :] = sandwhich, sandwhich*10
                                    # make ufloat, operate on it and split
                                    val = rvar_tot[imat,  iza, 0]
                                    val = np.sqrt(val).item(0)
                                    rsd_tot[imat, iza, :] = val, val*10
                                else:
                                    rvar_tot[imat, iza, :] = sandwhich.n, sandwhich.s
                                    # make ufloat, operate on it and split
                                    val = ufloat(rvar_tot[imat,  iza, 0],
                                                 rvar_tot[imat, iza, 1])

                                    val = unp.sqrt(val).item(0)
                                    rsd_tot[imat, iza, :] = val.n, val.s

                            else:
                                if noUQ is True:
                                    # make ufloat, operate on it and split
                                    val = rvar_tot[imat,  iza, 0]
                                    val = val+sandwhich
                                    rvar_tot[imat, iza, :] = val, val*10

                                    val = unp.sqrt(val).item(0)
                                    rsd_tot[imat, iza, :] = val, val*10
                                else:
                                    # make ufloat, operate on it and split
                                    val = ufloat(rvar_tot[imat,  iza, 0],
                                                 rvar_tot[imat, iza, 1])

                                    val = val+sandwhich
                                    rvar_tot[imat, iza, :] = val.n, val.s

                                    val = unp.sqrt(val).item(0)
                                    rsd_tot[imat, iza, :] = val.n, val.s

                # assign ndarray to response-wise dict
                respRVAR[temp][resp] = rvar
                respRSD[temp][resp] = rsd

                # assign total sensitivies
                respRVAR_tot[temp][resp] = rvar_tot
                respRSD_tot[temp][resp] = rsd_tot

        # return dictionaries
        self.sensperts = SensMap
        self.respRVAR = respRVAR
        self.respRSD = respRSD
        self.respRVAR_tot = respRVAR_tot
        self.respRSD_tot = respRSD_tot
        # return other attributes
        self.materials = {v: k for k, v in enumerate(mats)}
        self.zais = {str(v): k for k, v in enumerate(zais)}
        self.mts = {v: k for k, v in enumerate(mts)}

    def getInfo(perts):
        """Get mat, ZAID, MT and perturbation number from sens.perts list."""
        MAT = []
        ZAID = []
        MT = []
        PertMap = {}

        for n, p in enumerate(perts):

            try:
                mat, za, mt, nB = p.split("_")

                if len(za) < 6:
                    za = int(za+'0')

                elif len(za) == 6 and za.endswith('0'):
                    za = int(za)

                else:
                    raise ValueError('Check ZA in perturbation file!')

                mt = "mt %s xs" % mt.replace("MT", "")

                if mat not in PertMap.keys():
                    PertMap[mat] = {}

                if za not in PertMap[mat].keys():
                    PertMap[mat][za] = {}

                    if mt not in PertMap[mat][za].keys():
                        PertMap[mat][za][mt] = [n]

                    else:
                        PertMap[mat][za][mt].append(n)

                else:

                    if mt not in PertMap[mat][za].keys():
                        PertMap[mat][za][mt] = [n]

                    else:
                        PertMap[mat][za][mt].append(n)

                MAT.append(mat)
                MT.append(mt)
                ZAID.append(za)

            except ValueError:
                PertMap['tot'] = {'1': 0}

        # TODO: raise error if same nuclide same material have different MTs

        return list(set(MAT)), list(set(ZAID)), list(set(MT)), PertMap

    def get(self, attr, resp, mat, zai, temp, mt=None):
        """
        Get mean and standard deviation from XGPT object.

        Parameters
        ----------
        attr : string
            Object attribute. It may be 'respRVAR', 'respRSD', 'respRVAR_tot'
            or 'respRSD_tot'.
        mat : string
            Material identifier.
        zai : string
            Z-A identifier.
        temp : tuple
            Couple of fuel and coolant temperatures used to evaluate the
            responses.
        mt : string, optional
            MT reaction. This is only needed for 'respRVAR' and 'respRSD'
             attributes. The default is ``None``.

        Returns
        -------
        vals : numpy.array
            Mean and standard deviation required.
        """
        vals = []
        iMat = self.materials[mat]
        try:
            iZai = self.zais[zai]
        except KeyError:
            iZai = self.zais[str(zai)]

        if attr.endswith('_tot') is False:

            if mt is None:

                raise ValueError('Specify MT numbers!')

            else:

                iMT = self.mts[mt]

            for k in self.__dict__[attr][temp].keys():

                if k.startswith(resp):

                    vals.append(self.__dict__[attr][temp][k][iMat, iZai, iMT,
                                                             :])

        else:

            for k in self.__dict__[attr][temp].keys():

                if k.startswith(resp):

                    vals.append(self.__dict__[attr][temp][k][iMat, iZai, :])

        return np.asarray(vals, dtype=float)

    def getDirectSens(indsens, RObasis, ROmap, parameter, response, fineG,
                      coarseG, sigma=2, lethG=None, noUQ=True):
        """
        Compute direct contribution to sensitivity coefficients.

        Parameters
        ----------
        indsens : obj
            Indirect sensitivity coefficients object parsed with serpentTools.
        RObasis : TYPE
            DESCRIPTION.
        ROmap : TYPE
            DESCRIPTION.
        parameter : TYPE
            DESCRIPTION.
        response : TYPE
            DESCRIPTION.
        fineG : TYPE
            DESCRIPTION.
        coarseG : TYPE
            DESCRIPTION.

        Returns
        -------
        dirsens : obj
            Direct sensitivity coefficients object with same structure of the
            sensitivity reader objects parsed with serpentTools.

        """
        dirsens = deepcopy(indsens)
        materials, zais, mts, SensMap = ndXGPT.getInfo(indsens.perts)
        nMax = min(dirsens.nPert-1, RObasis.nBasis)
        for ir, resp in enumerate(dirsens.sensitivities.keys()):
            # set coeffs to zero. If no direct effect, don't overwrite
            S = np.zeros((dirsens.nPert, 2))
            for dirResp in parameter.keys():
                # look for direct response match
                if resp.startswith(dirResp):
                    # take group from string
                    G = int(resp[-1])

                    idg = ((fineG >= coarseG[G])
                           & (fineG < coarseG[G+1]))
                    # select integration points
                    idg = np.argwhere(idg == 1)
                    iS = idg[0][0]
                    iE = idg[-1][0]

                    tmp = response['%sC' % dirResp]
                    if noUQ is True:
                        R = tmp[0][G]
                    else:
                        R = ufloat(tmp[0][G], tmp[0][G]*tmp[1][G]*sigma)

                    for za in zais:
                        for mt in mts:
                            for mat in materials:
                                pardict = parameter[dirResp]
                                if za in pardict.keys() and mt in pardict[za].keys():
                                    for iB in range(0, nMax):
                                        # select basis function
                                        nS, nE = ROmap[za][mt]
                                        bf = RObasis.basis[nS:nE+1, iB]
                                        if noUQ is True:
                                            Rh = parameter[dirResp][za][mt][0]
                                        else:
                                            tmp = (parameter[dirResp][za][mt][0],
                                                   parameter[dirResp][za][mt][1])
                                            Rh = np2unp(tmp[0], tmp[1]*sigma)

                                        idx = SensMap[materials[0]][za][mt][iB]
                                        if lethG is not None:
                                            tmp = wIntegrate(Rh, lethG,
                                                             bf[iS:iE+1],
                                                             fineG[iS:iE+2])

                                            if tmp is None:
                                                msg = 'Basis function energy grid should have %d elements' % (len(fineG)+1)
                                                raise OSError(msg)
                                            else:
                                                tmp = tmp/R

                                        else:
                                            tmp = trapz(bf[iS:iE+1]*Rh[iS:iE+1],
                                                        x=fineG[iS:iE+1])/R
                                        if noUQ is True:
                                            S[idx, 0] = tmp
                                            S[idx, 1] = np.ones(tmp.shape)
                                        else:
                                            # store nominal value
                                            S[idx, 0] = tmp.n
                                            # store relative standard deviation
                                            S[idx, 1] = tmp.s/(sigma*abs(tmp.n))

                dirsens.energyIntegratedSens[resp][0, 0, :, :] = deepcopy(S)

        return dirsens

    def getSens(indsensT, dirsensT):
        """
        Get direct plus indirect sensitivities.

        Parameters
        ----------
        indsensT: object
            Dictionary with sensitivity objects created reading Serpent-2
            "_sens0.m" object with "serpentTools" module. The keys are the
            temperature couples used to evaluate the sensivity coefficients.
        matcomp: dict
            Dictionary matching the nuclides considered for the sensitivity
            study
                       to each material considered.
        sigma: confidence interval required for statistical error
                      propagation.

        Returns
        -------
        None
        """
        sens = deepcopy(indsensT)
        for temp in indsensT.keys():
            # get temperature couple
            indsens = indsensT[temp]
            # loop over materials
            for iMat in range(0, indsens.nMat):
                # loop over nuclides
                for iZai in range(0, indsens.nZai):
                    # loop over responses
                    for resp in indsens.energyIntegratedSens.keys():
                        # get energy integrated sensitivity coeffs
                        eintsens = indsens.energyIntegratedSens[resp]
                        indSn = eintsens[iMat, iZai, :, 0]
                        indSs = eintsens[iMat, iZai, :, 1]

                        eintdirsens = dirsensT[temp].energyIntegratedSens[resp]
                        dirSn = eintdirsens[iMat, iZai, :, 0]
                        dirSs = eintdirsens[iMat, iZai, :, 1]
                        # convert S into ufloat type
                        # indS = np2unp(indSn, 2*indSs)
                        # dirS = np2unp(dirSn, 2*dirSs)
                        S = indSn+dirSn
                        # ST = np.full((len(S), 2), 0)
                        # for i, s in enumerate(S):
                        #     ST[i, 0] = s.n
                        #     ST[i, 1] = s.s/abs(2*s.n) if s.n != 0 else 0
                        # assign sensitivities
                        es = sens[temp].energyIntegratedSens[resp]
                        es[iMat, iZai, :, 0] = S
                        es[iMat, iZai, :, 1] = S*0

        return sens


def np2unp(v, rsd):
    """
    Convert numpy.ndarray into equivalent uncertain numpy.ndarray.

    Parameters
    ----------
    v : numpy.array, list
        Array representing average values.
    rsd : numpy.array, list
        Array representing relative standard deviation values.

    Returns
    -------
    None.

    """
    if v.shape != rsd.shape:
        raise ValueError('Mean and rsd shapes are not consistent!')

    if len(v.shape) == 1:
        # get
        uv = unp.uarray(v, rsd*abs(v))

    return uv


def uncformat(data, std, fmtn=":.2u%S", fmts=None):
    """
    Convert data and std into formatted string for printing.

    Parameters
    ----------
    data : numpy.array, list
        Array of average values.
    std : numpy.array, list
        Array of standard deviations.
    fmt : string
        Format for printing data.
    fmts : string
        Format for standard deviation. Default is None, so format is the same
        of the data.

    Returns
    -------
    out : string
        data+/-std deviation

    """
    if len(data) != len(std):
        raise ValueError("Data and std dimension mismatch!")

    if std is None:
        try:
            # data are ufloat
            inptup = [(d.n, d.s) for d in data]
        except OSError:
            raise TypeError("If std is not provided, data must be ufloat type")
    else:
        inptup = list(zip(data, std))

    fmtn = '{%s}' % fmtn
    out = []
    outapp = out.append
    for tup in inptup:
        d, s = tup
        outapp(fmtn.format(ufloat(d, s)))
    return out


def wIntegrate(f, xf, w, xw):
    """
    Evaluate weighted integral on two different meshes.

    Parameters
    ----------
    f : TYPE
        DESCRIPTION.
    xf : TYPE
        DESCRIPTION.
    w : TYPE
        DESCRIPTION.
    xw : TYPE
        DESCRIPTION.

    Returns
    -------
    integral : TYPE
        DESCRIPTION.

    """
    M = len(w)
    if len(xw) <= M:
        return None
    xl, xu = xw.min(), xw.max()
    idg = ((xf >= xl) & (xf < xu))
    idg = np.argwhere(idg == 1)
    iS = idg[0][0]
    iE = idg[-1][0] if xf[-1] != xu else len(xf)-1
    I = interp1d(xw, np.append(w, w[-1]), kind='zero')
    integral = trapz(I(xf[iS:iE+1])*f[iS:iE+1], xf[iS:iE+1])

    return integral
