"""
Author: N. Abrate.

File: POD.py

Description: Class for generating the POD basis associated to the input
              snapshot matrix
"""
import os
import time as t
import h5py as h5
import numpy as np
import matplotlib.pyplot as plt
from numpy.linalg import eigh, svd, norm
from scipy.linalg import fractional_matrix_power as matpow
from scipy import sparse


class POD:
    """
    Class for generating the POD basis associated to the input snapshot matrix.

    Attributes
    ----------
    nParam: int
        Parameter space dimension
    nSnap: int
        Number of snapshots
    nBasis: int
        Number of basis functions
    basis: numpy.numpy.ndarray
        POD basis
    eigv: numpy.numpy.ndarray
        Eigenvalues associated to the POD basis
    energy: float
        POD-basis energy
    algo: str
        Algorithm chosen for the POD
    coeffs : numpy.ndarray
        POD coefficients on the reduced space
    EYtruncerr: numpy.ndarray
        Eckart-Young theoretical error bound (i.e. (nBasis+1)th eigenvalue)
    maxtruncerr: float
        Maximum truncation error

    Methods
    -------
    loadPOD(self, h5name):
        Define POD reading a file.
    BuildPOD(Y, truncerr, algo='svd'):
        Extract a truncated POD basis from a snapshot matrix.
    truncatePOD(eigenvals, truncerr):
        Truncate the POD basis to satisfying the POD energy requirement.
    showPODdecay(self, threshold=``None``):
        Show plot with POD basis eigenvalue decay.
    weightMat(W, Y, algo):
        Define weighting matrix.
    obj2h5(self, h5name, attrs=``None``):
        Store instance in .h5 format.
    """

    def __init__(self, snapshotmatrix, truncerr=None, algo='svd', snapmap=None,
                 W=None):
        """
        Initialise POD class.

        Parameters
        ----------
        snapshotmatrix: numpy.ndarray or str.
            Matrix with snapshots along its columns. If string, it is read from
            external file in format .hdf5.
        truncerr: float
            Maximum truncation error for the POD basis
        algo: string, optional
            POD solver. Default is SVD ('svd'), while other available
            algorithms are the EVD along parameter dimension ('evd1'),
            and EVD along snapshots dimension ('evd2')
        h5name: string
            Path to the HDF5 file storing the POD object to be loaded.
        snapmap: dict, optional
            Dictionary mapping the snapshot matrix columns.
            Default is ``None``.

        Returns
        -------
        ``None``
        """
        if isinstance(snapshotmatrix, (np.ndarray, list)):
            # store dimension size and snapshot number
            self.nParam, self.nSnap = snapshotmatrix.shape
            # build POD basis
            if truncerr is None:
                truncerr = 1e-5

            W, broadcastdot, algo = POD.weightMat(W, snapshotmatrix, algo)
            psi, eigv, energy, algo = POD.BuildPOD(snapshotmatrix, truncerr,
                                                   algo, W=W,
                                                   broadcastdot=broadcastdot)
            # assign attributes
            self.basis, self.eigv = psi, eigv
            self.energy, self.algo = energy, algo
            # assign POD coefficients
            if broadcastdot is True:
                WY = np.multiply(W[:, None], snapshotmatrix)
            else:
                WY = np.dot(W, snapshotmatrix)
            self.coeffs = np.dot((self.basis).T, WY)  # projection coefficients
            # Eckart-Young theoretical error bound
            trunc_ind = psi.shape[1]
            self.nBasis = trunc_ind
            self.EYtruncerr = eigv[trunc_ind-1]
            # evaluate truncation error
            snapshotmatrixTR = np.dot(self.basis, self.coeffs)  # truncate SVD
            # norm of residual between matrices
            myerr = 0
            self.maxtruncerr = 0
            for ic in range(0, snapshotmatrix.shape[1]):
                col = snapshotmatrix[:, ic]
                TRcol = snapshotmatrixTR[:, ic]
                myerr = norm(col-TRcol)/norm(col)
                if myerr > self.maxtruncerr:
                    self.maxtruncerr = myerr

            if snapmap is not None:
                self.snapmap = snapmap

        elif isinstance(snapshotmatrix, str):
            POD.loadPOD(self, snapshotmatrix)

        else:
            raise OSError("Unknown type for snapshotmatrix variable!")

    def loadPOD(self, h5name):
        """
        Define POD reading a file.

        Parameters
        ----------
        h5name: str
            HDF5 file name containing the POD.

        Returns
        -------
        ``None``
        """
        datapath = POD.__checkname__(h5name)
        fh5 = h5.File(datapath, "r")
        # store dimension size and snapshot number
        self.nParam = int(np.array(fh5.get("POD/nParam")))
        self.nSnap = int(np.array(fh5.get("POD/nSnap")))
        # assign attributes
        self.basis = np.array(fh5.get("POD/basis"))
        self.eigv = np.array(fh5.get("POD/eigv"))
        self.energy = float(np.array(fh5.get("POD/energy")))
        self.algo = str(np.array(fh5.get("POD/algo")))
        self.coeffs = np.array(fh5.get("POD/coeffs"))
        self.nBasis = int(np.array(fh5.get("POD/nBasis")))
        self.EYtruncerr = float(np.array(fh5.get("POD/EYtruncerr")))
        self.maxtruncerr = float(np.array(fh5.get("POD/maxtruncerr")))

    def BuildPOD(Y, truncerr, algo='svd', W=None, broadcastdot=False):
        """
        Extract a truncated POD basis from a snapshot matrix.

        Parameters
        ----------
        Y: numpy.ndarray
            Matrix with snapshots along its columns.
        truncerr: float
            Maximum truncation error for the POD basis
        algo: str, optional.
            POD solver. Default is SVD ('svd'), while other available
            algorithms are the EVD along parameter dimension ('evd1'),
            and EVD along snapshots dimension ('evd2')
        W: numpy.ndarray, optional
            Weighting matrix/vector. Default is ``None``.
        broadcastdot: bool, optional
            If True, the weighting procedure is performed broadcasting the
            product. Default is ``False``.

        Returns
        -------
        basis: numpy.ndarray
            Matrix containing basis vectors along its column
        eigenvals:
            POD eigenvalues
        energy: float
            POD energy
        algo: str
            algorithm used for POD
        """
        # message for the user
        print('Extracting the truncated POD basis... ', end='')

        # check weight
        if algo == 'svd':

            try:
                # SVD
                if broadcastdot is True:
                    WW = np.power(W, 1/2)
                    YW = np.multiply(WW[:, None], Y)  # weighting
                else:
                    YW = np.dot(matpow(W, 1/2), Y)  # weighting

                psi, sigma, phi = svd(YW)

                if broadcastdot is True:
                    WW = np.power(W, -1/2)
                    psi = np.multiply(WW[:, None], psi)  # weighting
                else:
                    psi = np.dot(matpow(W, -1/2), psi)  # inverse weighting

                # sort eigenvalues
                idx = sigma.argsort()[::-1]
                psi, sigma, phi = psi[:, idx], sigma[idx], phi[:, idx]
                # manipulate eigenvalues
                eigenvals = np.square(sigma)
                # truncate POD
                truncind, energy = POD.truncatePOD(eigenvals, truncerr)
                # compute basis
                basis = psi[:, :truncind]

            except MemoryError:
                m, n = Y.shape  # parameters number, snapshots number

                if n > m:  # choice less-consuming RAM algorithm
                    algo = 'evd1'

                else:
                    algo = 'evd2'

        if algo == 'evd1':
            # define snapshot correlation matrix
            if broadcastdot is True:
                WW = np.power(W, 1/2)
                YW = np.multiply(WW[:, None], Y)  # weighting
            else:
                YW = np.dot(matpow(W, 1/2), Y)  # weighting

            R = np.dot(YW, YW.T)
            # EVD
            lamb, psi = eigh(R)

            if broadcastdot is True:
                WW = np.power(W, -1/2)
                psi = np.multiply(WW[:, None], Y)  # weighting
            else:
                psi = np.dot(matpow(W, -1/2), psi)  # inverse weighting

            # sort eigenvalues
            idx = (lamb).argsort()[::-1]  # abs
            lamb, psi = lamb[idx], psi[:, idx]
            # manipulate eigenvalues
            eigenvals = (lamb)  # abs
            truncind, energy = POD.truncatePOD(eigenvals, truncerr)
            basis = psi[:, :truncind]

        elif algo == 'evd2':
            # define snapshot correlation matrix
            if broadcastdot is True:
                WY = np.multiply(W[:, None], Y)
            else:
                WY = np.dot(W, Y)

            K = np.dot(Y.T, WY)
            # EVD
            lamb, phi = eigh(K)
            # sort eigenvalues
            idx = (lamb).argsort()[::-1]  # abs
            lamb, phi = lamb[idx], phi[:, idx]
            # manipulate eigenvalues
            eigenvals = (lamb)  # abs
            psi = np.dot(Y, phi)
            # truncate POD
            truncind, energy = POD.truncatePOD(eigenvals, truncerr)
            # compute basis
            basis = np.divide(psi[:, :truncind], np.sqrt(eigenvals[:truncind]))

        print('DONE')  # message for the user

        return basis, eigenvals, energy, algo

    def weightMat(W, Y, algo):
        """
        Define weighting matrix.

        Parameters
        ----------
        W: numpy.ndarray
            Weighting matrix/vector.
        Y: numpy.ndarray
            Matrix with snapshots along its columns.

        Raises
        ------
        OSError
            'Weighting matrix is not squared!'
            'Weighting matrix is not squared!'

        Returns
        -------
        W: numpy.ndarray
            Weighting matrix/vector.
        broadcastdot: bool, optional.
            If True, the weighting procedure is performed broadcasting the
            product. Default is False.
        algo: str.
            POD algorithm. It is chosen according to parameter dimension ('evd1'),
            or snapshots dimension ('evd2')

        """
        broadcastdot = False
        if W is not None:
            if len(W.shape) == 1 or 1 in W.shape:  # is array
                if W.shape[0] > 1E6:
                    # array too big, broadcastdot operations
                    broadcastdot = True
                    if len(W.shape) != 1:
                        W = np.squeeze(W)
                else:
                    # convert array in identity matrix
                    W = sparse.diags(W)

            else:  # is matrix?
                n, m = W.shape
                if n != m:
                    raise OSError('Weighting matrix is not squared!')

                W[np.isnan(W)] = 0
                if np.allclose(Y.T, Y) is False:
                    raise OSError('Weighting matrix is not symmetric!')

        else:
            try:
                W = np.identity(Y.shape[0], dtype=np.int8)
            except MemoryError:
                m, n = Y.shape  # parameters number, snapshots number
                if m > 1E6:
                    broadcastdot = True
                    W = np.ones((Y.shape[0],), dtype=np.int8)
                else:
                    W = sparse.identity(Y.shape[0], dtype=np.int8)
                    W = W.tocsr()

                if n > m:  # choice less-consuming RAM algorithm
                    algo = 'evd1'

                else:
                    algo = 'evd2'

        return W, broadcastdot, algo

    def truncatePOD(eigenvals, truncerr):
        """
        Truncate the POD basis to satisfying the POD energy requirement.

        Parameters
        ----------
        eigenvals: numpy.ndarray
            Array containing sorted eigenvalues
        truncerr: float
            Maximum truncation error for the POD basis

        Returns
        -------
        truncind: int
            Number of POD basis vectors (truncated dimension)
        """
        # compute the energy associated to each eigenvalue
        energy = eigenvals / np.sum(eigenvals)

        # compute index satisfying energy requirement to truncate the basis
        totenergy = 0

        for i in range(len(energy)):  # loop over eigenvalues
            totenergy = totenergy+energy[i]  # compute total energy

            if 1-totenergy <= truncerr:  # check energy requirement
                truncind = i+1
                break

            else:
                truncind = len(energy)

        return truncind, totenergy

    def showPODdecay(self, threshold=None, ax=None, **kwargs):
        """
        Show plot with POD basis eigenvalue decay.

        Parameters
        ----------
        threshold : int
            maximum number of POD eigenvalues to be shown.

        Raises
        ------
        OSError
            -File extension is wrong. Only HDF5 can be parsed

        Returns
        -------
        fname : string
            Filename and proper extension.

        """
        # assign threshold
        threshold = self.nBasis if threshold is None else int(threshold)

        ax = ax or plt.gca()
        # plot
        ax.scatter(np.arange(1, threshold+1), self.eigv[0:threshold], **kwargs)
        ax.set_yscale('log')
        ax.grid(True, alpha=0.3)
        return ax

    def __checkname__(fname):
        """
        Check extension of filename and add if it is missed.

        Parameters
        ----------
        fname : str
            Input filename (optional extension).

        Raises
        ------
        OSError
            -File extension is wrong. Only HDF5 can be parsed

        Returns
        -------
        fname : string
            Filename and proper extension.

        """
        lst = fname.split(".")

        if len(lst) == 1:
            # add extension
            fname = fname+".hdf5"

        else:

            if lst[-1] != "hdf5":
                lst[-1] = "hdf5"
                fname = ".".join(lst)
                raise OSError("File extension wrong. Only HDF5 can be parsed!")

        return fname

    def obj2h5(self, h5name, attrs=None):
        """
        Store instance in .h5 format.

        Parameters
        ----------
        self : object/list[object]
            class object.
        h5name : string
            File where data are stored.
        attrs : dict, optional
            Dictionary attributes that can be assigned for further description.
            Default is ``None``.

        Returns
        -------
        ``None``
        """
        # check object list
        if type(self) != "list":
            self = [self]

        if type(attrs) != "list":
            attrs = [attrs]

        # create h5 file
        h5name = POD.__checkname__(h5name)
        fh5 = POD.__wopen__(h5name)

        if fh5 == -1:
            print("File not overwritten.")

        # profiling
        print("Writing object in hdf5 file...")
        tstart = t.time()

        # loop over object attributes
        for i, (ob, at) in enumerate(zip(self, attrs)):
            # store lines to be found in a list
            if at is not None:
                fid = at['grp_name']

            else:
                fid = 'grp%g' % i
            # create group with output id
            fh5.create_group(fid)

            # store dict as metadata
            for k, v in at.items():
                if k != 'grp_name':
                    fh5[fid].attrs[k] = v

            # loop over object attributes
            for k2, v2 in ob.__dict__.items():
                # loop over different object attributes
                fh5[fid].create_dataset(k2, data=v2)

        # TODO: save each item with its own type

        fh5.close()
        print("DONE \n")
        print("Elapsed time %f s" % (t.time()-tstart))

    def __wopen__(h5name):
        """
        Open the hdf5 file "h5name.hdf5" in "append" mode.

        Parameters
        ----------
        h5name : str
            File name

        Returns
        -------
        fh5 : object
            h5py object

        """
        if os.path.isfile(h5name):
            print("File exists. Overwriting?")
            ans = input()

        else:  # if file do not exist, it creates it
            ans = "create"

        if ans == "yes" or ans == "y":
            os.remove(h5name)
            # open file in write mode
            fh5 = h5.File(h5name, "a")

            return fh5

        elif ans == "create":
            # create file in append mode
            fh5 = h5.File(h5name, "a")

            return fh5

        else:  # if answer is not positive, nothing is done
            return -1
