"""
author: N. Abrate.

file: XGPTPOD.py

description: POD sub-class for Serpent-2 xGPT module.
"""
import numpy as np
from .POD import POD
from matplotlib.pyplot import gca
from numpy.linalg import svd, norm


class XGPTPOD(POD):
    """
    Class extending POD class for Serpent-2 xGPT mode pre-processing.

    Attributes
    ----------
    matcomp : dict
        Dict mapping the material composition with respect to perturbed
        nuclides
    PODmap : dict
        Dict mapping the POD basis functions with respect to zais and mt
        reactions
    zais : str
        Z-A+"0" of the nuclide analysed
    mt : List
        List of MT reactions included
    energygrid : numpy.ndarray
        Energy grid where cross sections or covariance are defined upon.

    Methods
    -------
    write_basfunc4xGPT(self, fmt='%.15e', header='lin-lin'):
        Write basis functions to .txt files for Serpent-2 xGPT

    """

    def __init__(self, zais, mts, materials, energygrid, Y=None, truncerr=None,
                 algo='svd', h5name=None, snapmap=None, W=None,
                 iscovariance=True):
        """
        Initialise class.

        Parameters
        ----------
        zais: str
            Z-A+"0" of the nuclide analysed.
        mts: list[int]
            List of MT reactions contained in snapshot/covariance matrix.
        material: str
            Material name whose composing nuclides XS are perturbed.
        energygrid: numpy.ndarray
            Energy grid where cross sections or covariance are defined upon.
        Y: numpy.ndarray. Default is ``None``.
            Matrix with snapshots along its columns.
        truncerr: float. Default is ``None``.
            Maximum truncation error for the POD basis
        algo: str. Default is 'svd'.
            POD solver. Default is SVD ('svd'), while other available
            algorithms are the EVD along parameter dimension ('evd1'),
            and EVD along snapshots dimension ('evd2')
        h5name: str. Default is ``None``.
            Path to the HDF5 file storing the POD object to be loaded.
        snapmap: dict. Default is ``None``.
            Dictionary mapping the snapshot matrix columns. Default is ``None``.

        Returns
        -------
        ``None``
        """
        if iscovariance is False:
            super(XGPTPOD, self).__init__(Y, truncerr, algo, snapmap, W)
        else:
            # store dimension size and snapshot number
            self.nParam, self.nSnap = Y.shape

            # build POD basis
            if truncerr is None:
                truncerr = 1e-5
            # compute SVD
            psi, sigma, phi = svd(Y)

            # sort eigenvalues
            idx = sigma.argsort()[::-1]
            psi, sigma, phi = psi[:, idx], sigma[idx], phi[:, idx]
            # truncate POD
            truncind, totenergy = super(XGPTPOD, XGPTPOD).truncatePOD(sigma,
                                                                      truncerr)
            # compute basis
            basis = psi[:, :truncind]
            # assign attributes
            self.basis, self.eigv = basis, sigma
            self.energy, self.algo = totenergy, algo
            self.coeffs = np.dot((basis).T, Y)  # projection coefficients
            # Eckart-Young theoretical error bound
            self.nBasis = truncind
            self.EYtruncerr = sigma[truncind-1]
            # evaluate truncation error
            YT = np.dot(self.basis, self.coeffs)  # truncate SVD
            # norm of residual between matrices
            myerr = 0
            self.maxtruncerr = 0
            for ic in range(0, Y.shape[1]):
                col = Y[:, ic]
                TRcol = YT[:, ic]
                myerr = norm(col-TRcol)/norm(col)
                if myerr > self.maxtruncerr:
                    self.maxtruncerr = myerr

            if snapmap is not None:
                self.snapmap = snapmap

            # inherit methods

        NP = len(energygrid)-1

        # make MT nomenclature consistent with serpentTools
        mts.sort()  # sort (ascending)
        perts = {}
        PODmap = {}
        PODmap[zais] = {}
        for imt, mt in enumerate(mts):
            if mt == 1:
                perts['total xs'] = imt
                PODmap[zais]['total xs'] = [0+NP*imt, NP-1+NP*imt]
            else:
                perts['mt %g xs' % mt] = imt
                PODmap[zais]['mt %g xs' % mt] = [0+NP*imt, NP-1+NP*imt]

        self.materials = materials
        self.PODmap = PODmap
        self.zais = zais
        self.mt = perts
        self.energygrid = energygrid

    def write_basfunc4xGPT(self, fmt='%.15e', header='lin-lin'):
        """
        Write basis functions to .txt files for Serpent-2 xGPT calculation.

        Parameters
        ----------
        fmt: str. Default is '%.15e'.
            file output format.
        mts: list[int]
            List of MT reactions contained in snapshot/covariance matrix.
        material: str. Default is 'lin-lin'.
            file header. The other option are 'log-log', 'log-lin', 'lin-log'.

        Returns
        -------
        ``None``
        """
        materials, zais = self.materials, self.zais

        for imt, mt in enumerate(self.mt):

            MT = int(mt.split(" ")[1])

            for iBF in range(0, self.nBasis):
                name = '%s_MT%g_%g' % (zais, MT, (iBF+1))
                fname = "%s.txt" % name

                with open(fname, 'w') as f:
                    f.write('%s \n' % header)
                    f.write('%g \n' % (len(self.energygrid)-1))

                    idx = self.PODmap[zais][mt]
                    iS, iE = idx[0], idx[1]
                    x = zip(self.energygrid, self.basis[iS:iE+1, iBF])

                    for idx, row in enumerate(x):
                        line = " ".join(fmt % value for value in row)
                        f.write(line + '\n')

                sname = 'XGPT_%s.txt' % (zais)

                with open(sname, 'a') as f:
                    f.write('sens pert custom' +
                            ' %s efunc "%s" zailist %s matlist %s mtlist %g \n'
                            % (name, fname, zais, " ".join(materials), MT))

    def plotPODbasis(self, which, zai=None, mt=None, normalize=True, ax=None,
                     mevscale=False, labelFmt=None, title=None, logx=True,
                     logy=False, loglog=False, xlabel=None, ylabel=None,
                     legend=None, ncol=1):
        """
        Plot basis functions.

        Parameters
        ----------
        which: int
            Basis function number to be plotted.
        zai: ``None`` or str or int or iterable
            Plot basis functions due to these isotopes. Passing ``None``
            will plot against all isotopes.
        mt: ``None`` or str or list of strings
            Plot basis functions due to these MT reactions. Passing ``None``
            will plot against all perturbations.
        {ax}
        labelFmt: ``None`` or str
            Formattable string to be applied to the labels.
            The following entries will be formatted for each plot
            permuation::

                {m} - name of the material
                {z} - isotope zai
                {p} - specific perturbation
                {r} - response being plotted

        {logx}
        {logy}
        {loglog}
        {xlabel}
        {ylabel}
        {legend}
        {ncol}

        Returns
        -------
        {rax}

        Raises
        ------
        KeyError
            If response or any passed perturbation settings are not
            present on the object

        See Also
        --------
        * :py:meth:`str.format` - used for formatting labels

        """
        labelFmt = labelFmt or "mat: {m} zai: {z} MT: {mt}"

        ax = ax or gca()

        iS, iE = self.snapmap[zai][mt]
        yvals = self.basis[iS:iE+1, which]
        xvals = self.energies * 1E6 if mevscale is False else self.energies

        label = labelFmt.format(z=zai, p=mt)
        ax.plot(xvals, yvals, label=label)

        if xlabel is None:
            if mevscale is False:
                ax.xlabel = 'Energy [eV]'
            else:
                ax.xlabel = 'Energy [MeV]'
        else:
            xlabel = xlabel

        ax.ylabel = 'Basis function'

        if loglog or logx is True:
            ax.xscale('log')

        if loglog or logx is True:
            ax.yscale('log')

        return ax

    def showPODdecay(self, threshold=None, ax=None, **kwargs):
        super().showPODdecay(threshold=threshold, **kwargs)

    def obj2h5(self, h5name, attrs=None):
        super(XGPTPOD, self).obj2h5(self, h5name, attrs=None)
